from pathlib import Path
from pylmkit.core.base import BaseKnowledgeBase
from typing import Any, Optional, Union, Sequence, Dict
from pylmkit.utils.data_utils import text_as_document, dict_as_document, document_as_dict


class Text2Document(BaseKnowledgeBase):
    """
    td = Text2Document(texts=['2222222', '5555555555'], metadatas=[{}, {}])
    a = td.load(mode='dict')
    print(a)
    """

    def __init__(self, init_documents=None):
        super().__init__(init_documents=init_documents)

    @classmethod
    def get(cls, texts, metadatas=None, types="Document", is_return=True, return_mode='doc', extend=True):
        data = text_as_document(texts=texts, metadatas=metadatas, types=types)
        cls()._base(documents=data, is_return=is_return, return_mode=return_mode, extend=extend)


class Dict2Document(BaseKnowledgeBase):
    def __init__(self, init_documents=None):
        super().__init__(init_documents=init_documents)

    @classmethod
    def get(cls, doc_dict, is_return=True, return_mode='doc', extend=True):
        data = dict_as_document(doc_dict)
        cls()._base(documents=data, is_return=is_return, return_mode=return_mode, extend=extend)


class BaseDocument(BaseKnowledgeBase):
    def __init__(self, init_documents=None):
        super().__init__(init_documents=init_documents)


class BaseSplitter(BaseKnowledgeBase):

    def __init__(self, splitter=None, chunk_size=500, chunk_overlap=100, **kwargs):
        if splitter is None:
            from pylmkit.perception.directory import RecursiveCharacterTextSplitter
            self.splitter = RecursiveCharacterTextSplitter(chunk_size=chunk_size, chunk_overlap=chunk_overlap, **kwargs)
        else:
            self.splitter = splitter

    def get_split_documents(self, documents, return_mode='doc'):
        self.splitter_documents.extend(self.splitter.split_documents(documents))
        self._base_return(documents=self.splitter_documents, return_mode=return_mode)

    def get_split_text(self, text, return_mode='doc'):
        self.splitter_documents.extend(self.splitter.split_text(text))
        self._base_return(documents=self.splitter_documents, return_mode=return_mode)

    def get_split_texts(self, texts, return_mode='doc'):
        for text in texts:
            self.splitter_documents.extend(self.splitter.split_text(text))
        self._base_return(documents=self.splitter_documents, return_mode=return_mode)

    def _base_return(self, documents, return_mode='doc'):
        if return_mode == 'doc':
            return documents
        else:
            return document_as_dict(documents)


class WebLoader(BaseKnowledgeBase):
    """
    x = WebLoader(web_path="https://www.espn.com/")
    a = x.load()
    x.save_splitter_documents('qqqqq.yaml', chunk_size=20, chunk_overlap=10)
    print(a)
    b = x.split(chunk_size=100, chunk_overlap=20)
    print(b)
    """

    def __init__(self,
                 path: Union[str, Sequence[str]] = "",
                 default_parser: str = "html.parser",
                 requests_kwargs: Optional[Dict[str, Any]] = None,
                 requests_per_second: int = 2,
                 raise_for_status: bool = False,
                 bs_get_text_kwargs: Optional[Dict[str, Any]] = None,
                 bs_kwargs: Optional[Dict[str, Any]] = None,
                 header_template: Optional[dict] = None,
                 init_documents=None,
                 **kwargs
                 ):
        from langchain.document_loaders import WebBaseLoader
        super().__init__(init_documents=init_documents)
        data = WebBaseLoader(web_path=path,
                             default_parser=default_parser,
                             requests_kwargs=requests_kwargs,
                             requests_per_second=requests_per_second,
                             raise_for_status=raise_for_status,
                             bs_get_text_kwargs=bs_get_text_kwargs,
                             bs_kwargs=bs_kwargs,
                             header_template=header_template,
                             **kwargs
                             ).load()
        self._base(documents=data, is_return=False, extend=True)

    def get(self,
            is_return=True,
            return_mode='doc',
            extend=True,
            ):
        result = self._base(documents=self.documents, is_return=is_return, return_mode=return_mode, extend=extend)
        if is_return:
            return result


class DocumentLoader(BaseKnowledgeBase):
    """
    dgl = DocumentGlobLoader(path='./',
                         # suffixes=['.txt', '.py', '.json'],  # 只加载指定格式的文件
                         show_progress=True,
                         silent_errors=True,  # 为 True 时，忽略报错
                         )
    a = dgl.load()
    print(len(a), a)
    b = dgl.split(chunk_size=100, chunk_overlap=50)
    print(len(b), b)
    """
    def __init__(self,
                 path: str,
                 loader_cls=None,
                 loader_kwargs: Union[dict, None] = None,
                 glob: str = "**/[!.]*",
                 suffixes: Optional[Sequence[str]] = None,
                 recursive: bool = False, show_progress: bool = True,
                 use_multithreading: bool = True,
                 max_concurrency: int = 10,
                 silent_errors: bool = True,
                 load_hidden: bool = False,
                 init_documents=None,
                 **kwargs
                 ):
        from pylmkit.perception.directory import DirectoryLoader
        from pylmkit.perception.directory import UnstructuredFileLoader
        super().__init__(init_documents=init_documents)
        if loader_kwargs is None:
            loader_kwargs = {}
        if Path(path).is_dir():  # is dir?
            data = DirectoryLoader(path=path,
                                   glob=glob,
                                   suffixes=suffixes,
                                   loader_cls=loader_cls,
                                   loader_kwargs=loader_kwargs,
                                   recursive=recursive,
                                   show_progress=show_progress,
                                   use_multithreading=use_multithreading,
                                   max_concurrency=max_concurrency,
                                   silent_errors=silent_errors,
                                   load_hidden=load_hidden,
                                   **kwargs
                                   ).load()
            self._base(documents=data, is_return=False, extend=True)
        else:
            if loader_cls is None:
                loader_cls = UnstructuredFileLoader
            data = loader_cls(str(path), **loader_kwargs).load()
            self._base(documents=data, is_return=False, extend=True)

    def get(self,
            is_return=True,
            return_mode='doc',
            extend=True,
            ):
        result = self._base(documents=self.documents, is_return=is_return, return_mode=return_mode, extend=extend)
        if is_return:
            return result





import requests, os
from tqdm import tqdm
from concurrent.futures import ThreadPoolExecutor


def split_text(text, chunk_size=3000):
    # 将文本分割成指定大小的段落
    return [text[i: i + chunk_size] for i in range(0, len(text), chunk_size)]


def merge_abstracts(abstracts):
    # 将多个摘要合并为一个字符串
    return ' '.join(abstracts)


def summary(text, worker, max_chunk_size: int = 3000, max_summary_size: int = 500, show_progress: bool = True,
            max_workers: int = 5):
    """支持短文本、长文本摘要提取。

        ## 输入参数
        - text，str，需要提取摘要的文本
        - worker，function，大模型提取摘要函数
        - max_chunk_size，int，最大文本块，当大于最大长度时，将采用分段提取摘要，然后在汇总摘要，默认取 3000，
        - max_summary_size，int，最大摘要长度，当汇总后的摘要长度大于最大长度时，将采用分段提取摘要，然后在汇总摘要，默认取 500，
        - show_progress，bool，是否显示进度条，默认取 True，
        - max_workers，int，最大线程数，默认取 5，

        ## 输出参数
        - combined_abstract，str，摘要
    """
    text = text if isinstance(text, str) else text.content
    # 分段提取摘要
    chunks = split_text(text, chunk_size=max_chunk_size)
    # 用于存储摘要的列表
    abstracts = []
    # 使用多线程并发处理每个段的摘要生成
    # with ThreadPoolExecutor(max_workers=5) as executor:  # 根据实际情况调整线程数量
    #     futures = [executor.submit(worker, chunk) for chunk in chunks]
    #     abstracts = [future.result() for future in futures]
    with ThreadPoolExecutor(max_workers=max_workers) as executor:  # 根据实际情况调整线程数量
        if show_progress:
            # 使用tqdm显示进度
            for future in tqdm(executor.map(worker, chunks), total=len(chunks), desc="extract summary ..."):
                abstracts.append(future)
        else:
            # 不显示进度
            for future in executor.map(worker, chunks):
                abstracts.append(future)
    # 合并摘要
    combined_abstract = merge_abstracts(abstracts)
    # 如果合并后的摘要长度大于 max_summary_size，则递归调用函数
    if len(combined_abstract) > max_summary_size:
        return summary(combined_abstract,
                       worker=worker,
                       max_chunk_size=max_chunk_size,
                       max_summary_size=max_summary_size,
                       show_progress=show_progress,
                       max_workers=max_workers
                       )
    else:
        combined_abstract = worker(combined_abstract)
        return combined_abstract


# 主函数
def batch_summary(texts: list, worker, max_chunk_size: int = 3000, max_summary_size: int = 500, show_progress: bool = True,
                  max_workers: int = 5):
    """支持批量短文本、长文本摘要提取。

        ## 输入参数
        - text，list，需要提取摘要的多个文本
        - worker，function，大模型提取摘要函数
        - max_chunk_size，int，最大文本块，当大于最大长度时，将采用分段提取摘要，然后在汇总摘要，默认取 3000，
        - max_summary_size，int，最大摘要长度，当汇总后的摘要长度大于最大长度时，将采用分段提取摘要，然后在汇总摘要，默认取 500，
        - show_progress，bool，是否显示进度条，默认取 True，
        - max_workers，int，最大线程数，默认取 5，

        ## 输出参数
        - abstracts，list[str]，摘要
    """
    abstracts = []
    texts = [text if isinstance(text, str) else text.content for text in texts]
    if show_progress:
        for text in tqdm(texts, desc="extract batch summary ..."):
            combined_abstract = summary(text,
                                        worker=worker,
                                        max_chunk_size=max_chunk_size,
                                        max_summary_size=max_summary_size,
                                        show_progress=show_progress,
                                        max_workers=max_workers
                                        )
            abstracts.append(combined_abstract)
    else:
        for text in texts:
            combined_abstract = summary(text,
                                        worker=worker,
                                        max_chunk_size=max_chunk_size,
                                        max_summary_size=max_summary_size,
                                        show_progress=show_progress,
                                        max_workers=max_workers
                                        )
            abstracts.append(combined_abstract)
    return abstracts













def input_prompt(**kwargs):
    return kwargs


def return_language(language='English'):
    return f"\nReply in {language}"


def get_summary_default_prompt():
    _prompt = "提取下面内容的摘要：\n\ncontent: {content}"
    return _prompt










from pylmkit.app import RolePlay
from pylmkit.core.base import BaseWebUI


class RAGWebUI(BaseWebUI):
    def __init__(self,
                 title=None,
                 page_icon=None,
                 layout="centered",
                 language='en',
                 sidebar_title=None,
                 sidebar_describe=None,
                 footer_describe=None,
                 logo1=None,
                 logo2=None,
                 greetings=None,
                 placeholder=None,
                 refer_name=None,
                 ):
        super().__init__(
            title=title,
            page_icon=page_icon,
            layout=layout,
            language=language,
            sidebar_title=sidebar_title,
            sidebar_describe=sidebar_describe,
            footer_describe=footer_describe,
            logo1=logo1,
            logo2=logo2,
            greetings=greetings,
            placeholder=placeholder,
            refer_name=refer_name
        )





from pylmkit.utils.data_utils import stream_print


class LocalLLMModel(object):
    def __init__(self, model_path, language="en", tokenizer_kwargs=None, model_kwargs=None):
        import torch, logging
        import transformers
        self.history = []
        if tokenizer_kwargs is None:
            tokenizer_kwargs = {}
        if model_kwargs is None:
            model_kwargs = {}
        if language in ['zh']:
            try:
                import modelscope
            except:
                raise Exception("Please install `pip install modelscope`")
            self.tokenizer = modelscope.AutoTokenizer.from_pretrained(model_path,
                                                                      trust_remote_code=True,
                                                                      **tokenizer_kwargs
                                                                      )
            self.model = modelscope.AutoModelForCausalLM.from_pretrained(model_path,
                                                                         device_map="auto",
                                                                         offload_folder=model_path,
                                                                         trust_remote_code=True,
                                                                         **model_kwargs
                                                                         )
        else:
            self.tokenizer = transformers.AutoTokenizer.from_pretrained(model_path,
                                                                        trust_remote_code=True,
                                                                        **tokenizer_kwargs
                                                                        )
            self.model = transformers.AutoModelForCausalLM.from_pretrained(model_path,
                                                                           trust_remote_code=True,
                                                                           **model_kwargs
                                                                           )

        if torch.cuda.is_available():  # GPU
            # print("CUDA")
            self.model = self.model.half().cuda()
        else:  # CPU
            # print("CPU")
            self.model = self.model.float()
        self.model = self.model.eval()

    def invoke(self, query, history=None, **kwargs):
        # system=""  system相当于一个预设角色和引导作用
        response, self.history = self.model.chat(self.tokenizer, query, history=history, **kwargs)
        return response

    def stream(self, query, history=None, buffer_size=3, **kwargs):
        response, self.history = self.model.chat(self.tokenizer, query, history=history, **kwargs)
        return stream_print(response, buffer_size=buffer_size)






from pylmkit.core.prompt import input_prompt

init_css = '''
<style>
    #MainMenu {visibility: hidden;}
    header {visibility: hidden;}
    footer {visibility: hidden;}
    .css-z5fcl4 {
        width: 100%;
        padding: 1rem 1rem 1rem;  # 默认 6 1 10，改成1 1 1，修改页面顶部的空白
        padding-top: 0rem;  // 默认6，改成0，修改页面顶部的空白
        padding-right: 2rem;
        padding-bottom: 10rem;
        padding-left: 1rem;
        min-width: auto;
        max-width: initial;
    }
    #logo1 {
                position: fixed;
                display: flex;
                right: 40px;
                top: 20px;
                align-items: center;
            }
    #logo1 img {
      width: 30px;
      margin-right: 10px;
      border-radius: 50%; /* 添加 border-radius 属性 */
    }
    #logo2 {
                    position: fixed;
                    display: flex;
                    right: 80px;
                    top: 20px;
                    align-items: center;
                }
    #logo2 img {
        width: 30px;
        margin-right: 80px;
        border-radius: 50%;
    }
    /* 这个选择器更具体，可以覆盖Streamlit默认样式 */
    footer{
      visibility: hidden;
    }
    /* 在新的div标签中添加你自己的内容 */
    div.my-footer {
      position: fixed;
      bottom: 0;
      left: 0;
      width: 100%;
      height: 50px;
      display: flex;
      align-items: center;
      justify-content: center;
      color: #fff;
      background-color: #333;
    }

</style>
'''

init_footer = '''
<div class="my-footer">{}</div>
'''

init_logo = '''
<div id="{logo_id}">
    <a href="{link}">
        <img src="{logo_rul}" />
    </a>
    <span>{logo_title}</span>
</div>
'''

_zh = input_prompt(
    _title="PyLMKit",
    _page_icon="😄",
    _sidebar_title="😄 PyLMKit",
    _refer_name="引用信息",
    _greetings="您好，我能为您做些什么?",
    _placeholder="请输入您的消息...",
    _footer_describe="Copyright © 2023 PyLMKit  |  Make with Streamlit",
    _sidebar_describe='''
**pylmkit**: 帮助用户快速构建实用的大模型应用! [pylmkit](https://github.com/52phm/pylmkit)

- 开源协议 <br>Apache License 2  [detail](https://www.apache.org/licenses/LICENSE-2.0)
- 下载安装
```bash
pip install pylmkit -U
```
- 学习教程<br>
    - [English document](http://en.pylmkit.cn) 
    - [中文文档](http://zh.pylmkit.cn)
''',
    _logo1=input_prompt(
        logo_id="logo1",
        link="http://app.pylmkit.cn",
        logo_rul="https://img1.baidu.com/it/u=2672705872,739783853&fm=253&fmt=auto&app=138&f=JPEG?w=500&h=500",
        logo_title="pylmkit"
    ),
    _logo2=input_prompt(
        logo_id="logo2",
        link="https://github.com/52phm/pylmkit",
        logo_rul="https://github.githubassets.com/favicons/favicon.svg",
        logo_title=""
    ),

)

_en = input_prompt(
    _title="PyLMKit",
    _page_icon="😄",
    _sidebar_title="😄 PyLMKit",
    _greetings="How can I help you?",
    _placeholder="Your message...",
    _refer_name="Citation Information",
    _footer_describe="Copyright © 2023 PyLMKit  |  Make with Streamlit",
    _sidebar_describe='''
**pylmkit**: Help users quickly build practical large model applications! [pylmkit](https://github.com/52phm/pylmkit)

- LICENSE <br>Apache License 2  [detail](https://www.apache.org/licenses/LICENSE-2.0)
- How to install?
```bash
pip install pylmkit -U
```
- How to use?<br>
    - [English document](http://en.pylmkit.cn) 
    - [中文文档](http://zh.pylmkit.cn)
''',
    _logo1=input_prompt(
        logo_id="logo1",
        link="http://app.pylmkit.cn",
        logo_rul="https://img1.baidu.com/it/u=2672705872,739783853&fm=253&fmt=auto&app=138&f=JPEG?w=500&h=500",
        logo_title="pylmkit"
    ),
    _logo2=input_prompt(
        logo_id="logo2",
        link="https://github.com/52phm/pylmkit",
        logo_rul="https://github.githubassets.com/favicons/favicon.svg",
        logo_title=""
    ),

)



# main.py
from dotenv import load_dotenv
from pylmkit import BaseWebUI
from pylmkit.llms import ChatQianfan
from pylmkit.llms import EmbeddingsHuggingFace
from langchain.vectorstores import FAISS
from pylmkit.perception.text import DocumentLoader
from pylmkit.perception.text import WebLoader
from pylmkit.app import DocRAG


load_dotenv()
web = BaseWebUI(language='zh')

# load: documents
# loader = DocumentLoader(path='./document_test/aaa.txt')  # 加载一个文档
# loader = DocumentLoader(path='./document_test', show_progress=False)  # 批量加载
# docs = loader.split(chunk_size=200, chunk_overlap=50)
# docs = loader.get()
# print(len(docs))
# print(docs[-1])

# load: web
loader = WebLoader(path='https://zhuanlan.zhihu.com/p/339971541')
docs = loader.split(
    chunk_size=web.param(label="数据块大小", type='int', value=200),
    chunk_overlap=web.param(label="数据块重叠大小", type='int', value=50),
)
# docs = loader.split(
#     chunk_size=200,
#     chunk_overlap=50,
# )

# rag
embed_model = EmbeddingsHuggingFace(model_name="all-MiniLM-L6-v2")  #
vdb_model = FAISS
llm_model = ChatQianfan(model="ERNIE-Bot-turbo")
role_template = "{ra}\n user question: {query}"
rag = DocRAG(
    embed_model=embed_model,
    vdb_model=vdb_model,
    llm_model=llm_model,
    corpus=docs,
    role_template=role_template,
    return_language="中文",
    online_search_kwargs={},
    # online_search_kwargs={'topk': 2, 'timeout': 20},  # 搜索引擎配置，不开启则可以设置为 online_search_kwargs={}
)

web.run(
    obj=rag.invoke,
    input_param=[{"name": "query", "label": "用户输入内容", "type": "chat"},
                 {"name": "topk", "label": "最相似topk", "type": "int", "value": 5}
                 ],
    output_param=[{'label': '结果', 'name': 'ai', 'type': 'chat'},
                  {'label': '参考', 'name': 'refer', 'type': 'refer'}
                  ]
)

# rag.invoke(query="怎样评价宁德时代？")





from abc import ABC
import time
import pandas as pd
from pathlib import Path
from tqdm import tqdm
import streamlit as st
from pylmkit.utils.data_utils import read_yaml, read_json, write_yaml, write_json
from pylmkit.utils.data_utils import message_as_string, document_as_dict, dict_as_document
from pylmkit.utils.data_utils import text_as_document
from functools import partial
from pylmkit.core.html import init_css, init_footer, init_logo
from pylmkit.core.html import _zh, _en


class BaseMemory(object):
    human_prefix: str = "Human"
    ai_prefix: str = "AI"
    system_prefix: str = "System"

    def __init__(self, init_memory=None, streamlit_web=False):
        self.memory_messages = []
        self.streamlit_web = streamlit_web
        if self.streamlit_web:  # streamlit rerun page, so need cache
            if "memory" not in st.session_state:
                st.session_state["memory"] = self.memory_messages
        if isinstance(init_memory, list):
            self.memory_messages = init_memory
            if self.streamlit_web:
                st.session_state['memory'] = self.memory_messages
        if self.streamlit_web:  # streamlit rerun page, so need cache
            self.memory_messages = st.session_state['memory']

    def add(self, role, content, refer=''):
        """ role，human ai system
        """
        if role in ['user', 'User', 'USER', 'human', 'Human', 'HUMAN']:
            role = self.human_prefix
        elif role in ['ai', 'Ai', 'AI', 'assistant']:
            role = self.ai_prefix
        elif role in ['sys', 'system', 'System', 'SYS', 'SYSTEM']:
            role = self.system_prefix
        else:
            raise Exception(f"The role `{role}` does not exist")
        self.memory_messages.append(
            {"role": role, "content": content, "refer": refer, "date": time.strftime('%Y-%m-%d %H:%M:%S')})
        if self.streamlit_web:  # streamlit rerun page, so need cache
            st.session_state['memory'] = self.memory_messages

    def to_csv(self, filepath, index=False, **kwargs):
        data = self.memory_messages
        pd.DataFrame(data).to_csv(filepath, index=index, **kwargs)

    def clear(self):
        self.memory_messages = []
        if self.streamlit_web:  # streamlit rerun page, so need cache
            st.session_state['memory'] = self.memory_messages

    def _get(self, mode='message'):
        if mode == 'message':
            return self.memory_messages
        elif mode == 'string':
            return message_as_string(self.memory_messages)
        else:
            raise Exception(f"There is no such `{mode}` mode. Support modes: message, string")


class BaseKnowledgeBase(object):
    def __init__(self, init_documents=None):
        self.documents = []
        self.splitter_documents = []
        if isinstance(init_documents, list):
            self.documents = init_documents

    @classmethod
    def load(cls, filepath, is_return=True, return_mode="doc", extend=True):
        if filepath.endswith('.json'):
            data = read_json(filepath)
        elif filepath.endswith('.yaml') or filepath.endswith('yml'):
            data = read_yaml(filepath)  # data=[{},{}]
        else:
            raise Exception(f"The file type is not supported")
        data_dict_as_document = dict_as_document(data)
        result = cls()._base(documents=data_dict_as_document, return_mode=return_mode, is_return=is_return,
                             extend=extend)
        if is_return:
            return result

    @classmethod
    def add(cls, texts, metadatas=None, is_return=True, return_mode="doc", extend=True, types="Document"):
        data_dict_as_document = text_as_document(texts=texts, metadatas=metadatas, types=types)
        result = cls()._base(documents=data_dict_as_document, return_mode=return_mode, is_return=is_return,
                             extend=extend)
        if is_return:
            return result

    def split(self, splitter=None, chunk_size=500, chunk_overlap=100, return_mode='doc', **kwargs):
        if splitter is None:
            from pylmkit.perception.directory import RecursiveCharacterTextSplitter
            splitter = RecursiveCharacterTextSplitter(chunk_size=chunk_size, chunk_overlap=chunk_overlap, **kwargs)
        else:
            splitter = splitter
        self.splitter_documents = splitter.split_documents(self.documents)
        if return_mode == 'doc':
            return self.splitter_documents
        else:
            return document_as_dict(self.splitter_documents)

    def to_csv_loader(self, filepath, index=False, **kwargs):
        data = document_as_dict(self.documents)
        pd.DataFrame(data).to_csv(filepath, index=index, **kwargs)

    def to_csv_splitter(self,
                        filepath,
                        splitter=None,
                        chunk_size=500,
                        chunk_overlap=100,
                        index=False,
                        splitter_kwargs={},
                        csv_kwargs={}
                        ):
        if not self.splitter_documents:
            self.splitter_documents = self.split(splitter=splitter, chunk_size=chunk_size,
                                                 chunk_overlap=chunk_overlap, **splitter_kwargs)
        data = document_as_dict(self.splitter_documents)
        pd.DataFrame(data).to_csv(filepath, index=index, **csv_kwargs)

    def clear(self, mode='doc'):
        if mode == 'doc':
            self.documents = []
        else:
            self.splitter_documents = []

    def _base(self, documents, is_return=True, return_mode='doc', extend=True):
        if extend:
            self.documents.extend(documents)  # # dict -> Document
            if is_return:
                if return_mode == 'doc':
                    return self.documents
                else:
                    return document_as_dict(self.documents)
        else:
            # self.documents = documents  # when extend is False, just reset documents
            if is_return:
                if return_mode == 'doc':
                    return documents
                else:
                    return document_as_dict(documents)


def input_widget(input1, input2, type, value):
    if type == "int":
        return st.number_input(format='%d', step=1, **input1)
    if type == "float":
        return st.number_input(format='%f', **input1)
    elif type in ['list', 'List', 'select']:
        return st.selectbox(options=value, **input2)
    elif type == "bool":
        if value in [True, 'True', 'true']:
            options = [True, False]
        else:
            options = [False, True]
        return st.radio(options=options, horizontal=True, **input2)
    elif type == "file":
        uploaded_file = st.file_uploader(**input2)
        if uploaded_file is not None:
            res = str(Path().cwd() / uploaded_file.name)
            with open(res, "wb") as f:
                f.write(uploaded_file.getbuffer())
        else:
            res = None
        return res
    elif type in ['multiselect']:
        return st.multiselect(options=value, **input2)
    else:
        return st.text_input(**input1)


def generate_input_widget(mode="main", **kwargs):  # 在前端生成输入框
    """
    mode, default "main" ,other "sidebar"
    """
    label = kwargs.get('label', "")
    value = kwargs.get('value', None)
    name = kwargs.get('name', None)
    _input1 = {"label": label, "value": value, "key": f"{name}-{label}"}
    _input2 = {"label": label, "key": f"{name}-{label}"}
    _type = kwargs.get('type', None)  # int float bool string chat file
    if mode == 'main':
        return input_widget(
            input1=_input1,
            input2=_input2,
            type=_type,
            value=value
        )
    else:
        with st.sidebar:
            return input_widget(
                input1=_input1,
                input2=_input2,
                type=_type,
                value=value
            )


class BaseWebUI(object):
    def __init__(self,
                 title=None,
                 page_icon=None,
                 layout="centered",
                 language='en',
                 sidebar_title=None,
                 sidebar_describe=None,
                 footer_describe=None,
                 logo1=None,
                 logo2=None,
                 greetings=None,
                 placeholder=None,
                 refer_name=None,
                 ):
        self.title = title
        self.layout = layout
        self.page_icon = page_icon
        self.footer_describe = footer_describe
        self.sidebar_title = sidebar_title
        self.sidebar_describe = sidebar_describe
        self.logo1 = logo1
        self.logo2 = logo2
        self.greetings = greetings
        self.placeholder = placeholder
        self.refer_name = refer_name
        if language in ['zh', '中国', 'china']:
            self.lang = _zh
        else:
            self.lang = _en
        if self.title is None:
            self.title = self.lang.get('_title', '')
        if self.page_icon is None:
            self.page_icon = self.lang.get('_page_icon', None)
        if self.footer_describe is None:
            self.footer_describe = self.lang.get('_footer_describe', '')
        if self.sidebar_title is None:
            self.sidebar_title = self.lang.get('_sidebar_title', '')
        if self.sidebar_describe is None:
            self.sidebar_describe = self.lang.get('_sidebar_describe', '')
        if self.logo1 is None:
            self.logo1 = self.lang.get('_logo1', '')
        if self.logo2 is None:
            self.logo2 = self.lang.get('_logo2', '')
        if self.greetings is None:
            self.greetings = self.lang.get('_greetings', '')
        if self.placeholder is None:
            self.placeholder = self.lang.get('_placeholder', '')
        if self.refer_name is None:
            self.refer_name = self.lang.get('_refer_name', 'refer')

        self.base_page()
        if "messages" not in st.session_state:
            st.session_state["messages"] = [{"role": "assistant", "content": self.greetings}]
        self.input_kwargs = {}
        st.session_state["output_kwargs"] = {}
        for msg in st.session_state.messages:
            st.chat_message(msg["role"]).write(msg["content"])
            # refer setting
            refer = msg.get("refer", False)
            if refer:
                with st.expander(label=self.refer_name, expanded=False):
                    st.markdown(refer, unsafe_allow_html=True)

    def _input(self, content, role="user", avatar="😄"):
        st.chat_message(role, avatar=avatar).write(content, unsafe_allow_html=True)
        msg = {"role": role, "content": content}
        st.session_state.messages.append(msg)

    def _output(self, content, refer=None, role="assistant"):
        # st.chat_message(role).write(content, unsafe_allow_html=True)
        with st.chat_message(role):
            content_placeholder = st.empty()
            full_content = ""
            for chunk in content:
                full_content += chunk + ""
                time.sleep(0.01)
                content_placeholder.markdown(full_content + "▌")
            content_placeholder.markdown(full_content)

        if refer:  # refer setting
            with st.expander(label=self.refer_name, expanded=False):
                st.markdown(refer, unsafe_allow_html=True)
        msg = {"role": role, "content": content, "refer": refer}
        st.session_state.messages.append(msg)

    def output_parse(self, output_param, output_result):
        refer = None
        if len(output_param) == 0:
            response = None
        elif len(output_param) == 1:
            response = output_result
            st.session_state["output_kwargs"][output_param[0]['name']] = response
        else:
            response = output_result[0]
            for i, arg in enumerate(output_param):
                st.session_state["output_kwargs"][arg['name']] = output_result[i]
                if arg['type'] == 'chat':
                    response = output_result[i]
                if arg['type'] == 'refer':
                    refer = output_result[i]
        return response, refer

    def run(self, obj, input_param: list, output_param: list):
        chat_variable = ""
        obj = self.wrapper(obj)
        for arg in input_param:
            if arg['type'] != 'chat':
                self.input_kwargs[arg['name']] = generate_input_widget(mode='sidebar', **arg)
            else:
                chat_variable = arg['name']
        if chat_variable:
            if prompt := st.chat_input(placeholder=self.placeholder):
                self.input_kwargs[chat_variable] = prompt
                self._input(content=prompt)
                with st.spinner('PyLMKit: Generating, please wait...'):  # 正在生成，请稍候...
                    result = obj(**self.input_kwargs)
                    response, refer = self.output_parse(output_param, result)
                    self._output(content=response, refer=refer)
        else:
            with st.spinner('PyLMKit: Generating, please wait...'):  # 正在生成，请稍候...
                result = obj(**self.input_kwargs)
                response, refer = self.output_parse(output_param, result)
                # self._output(content=response, refer=refer)
                with st.expander(label="output", expanded=True):
                    st.json(st.session_state["output_kwargs"], expanded=True)

    def wrapper(self, fun):
        return partial(fun)

    def param(self, label, type, value, mode='sidebar'):
        input_kwargs = {
            "label": label,
            "type": type,
            "value": value
        }
        key = f"{label}-{type}-{str(value)}"
        if key not in st.session_state.keys():
            st.session_state[key] = ""
        renew_value = generate_input_widget(
            mode=mode,
            **input_kwargs
        )
        return renew_value

    def base_page(self):
        st.set_page_config(
            page_title=self.title,
            layout=self.layout,
            page_icon=self.page_icon,
        )
        st.markdown(init_css, unsafe_allow_html=True)
        if self.footer_describe:
            st.sidebar.markdown(init_footer.format(self.footer_describe), unsafe_allow_html=True)
        if self.sidebar_title:
            st.sidebar.title(self.sidebar_title)
        if self.sidebar_describe:
            st.sidebar.markdown(self.sidebar_describe, unsafe_allow_html=True)
        if self.logo2:
            st.markdown(init_logo.format(**self.logo2), unsafe_allow_html=True)
        if self.logo1:
            st.markdown(init_logo.format(**self.logo1), unsafe_allow_html=True)





import json
import toml
import pandas as pd
import yaml
from pathlib import Path
from yaml.loader import SafeLoader
from pydantic import Field, BaseModel
from typing import Literal
# from langchain.docstore.document import Document


def yield_specify_file(paths, suffixes):
    """Returns the file of the specified suffix"""
    for path in paths:
        if path.is_file():
            if suffixes and path.suffix not in suffixes:
                continue
            yield path


class Document(BaseModel):
    page_content: str
    metadata: dict = Field(default_factory=dict)
    type: str = "Document"

    def __str__(self):
        return f"Document(page_content='{self.page_content}', metadata={self.metadata})"


def stream_print(string, buffer_size=3):
    buffer = []
    for char in string:
        buffer.append(char)
        if len(buffer) >= buffer_size:
            yield ''.join(buffer)
            buffer.clear()
    if buffer:
        yield ''.join(buffer)


def document_as_refer(documents):
    document_refer = [f"[{i+1}] **{doc.metadata['source']}**  {doc.page_content}\n\n" for i, doc in enumerate(documents)]
    return "".join(document_refer)


def document_as_dict(documents):
    document_dict = [{"page_content": doc.page_content, "metadata": doc.metadata,
                      "type": doc.type} for doc in documents]
    return document_dict


def dict_as_document(doc_dict):
    document_dict = [Document(page_content=doc['page_content'],
                              metadata=doc.get('metadata', {}),
                              type=doc.get('type', 'Document')) for doc in doc_dict]
    return document_dict


def document_as_string(documents, sep=""):
    document_string = f"{sep}".join([doc.page_content for doc in documents])
    return document_string


def message_as_string(memory_messages):
    messages_string = [f"\n{message['role']}: {message['content']}" for message in memory_messages]
    return "".join(messages_string)


def text_as_document(texts, metadatas=None, types="Document"):
    documents = []
    if metadatas:
        if isinstance(types, str):
            for i, text in enumerate(texts):
                documents.append(Document(page_content=text, metadata=metadatas[i], type=types))
        else:  # types is `list` mode
            for i, text in enumerate(texts):
                documents.append(Document(page_content=text, metadata=metadatas[i], type=types[i]))
    else:
        if isinstance(types, str):
            for i, text in enumerate(texts):
                documents.append(Document(page_content=text, type=types))
        else:  # types is `list` mode
            for i, text in enumerate(texts):
                documents.append(Document(page_content=text, type=types[i]))
    return documents


def read_yaml(filepath):
    try:
        with open(filepath, encoding="utf-8") as fp:
            result = yaml.load(fp, Loader=SafeLoader)
    except Exception as e:
        raise Exception(e)
    return result


def write_yaml(data, filepath, mode="w", encoding='utf-8'):
    try:
        with open(filepath, mode=mode, encoding=encoding) as f:
            yaml.dump(data=data, stream=f, allow_unicode=True)
    except Exception as e:
        raise Exception(e)


def create_yaml(filepath, data):
    if not Path(filepath).exists():
        write_yaml(filepath, data)


def write_json(data, filepath, mode='w', encoding='utf-8', ensure_ascii=False):
    with open(filepath, mode, encoding=encoding) as fp:
        json.dump(data,  # 字典数据
                  fp=fp,  # open 文件
                  ensure_ascii=ensure_ascii,  # 确保中文无乱码
                  )


def read_json(filepath, mode='r', encoding='utf-8'):
    with open(filepath, mode, encoding=encoding) as fp:
        data = json.load(fp)
    return data


def read_csv_document(filepath, **kwargs):
    df = pd.read_csv(filepath, **kwargs)
    data = df.to_dict(orient='records')  # list(dict)
    return data


def read_csv_memory(filepath, **kwargs):
    df = pd.read_csv(filepath, **kwargs)
    data = df.to_dict(orient='records')  # list(dict)
    return data


def read_toml(file_path):
    with open(file_path, "r") as file:
        toml_data = toml.load(file)
    return toml_data


def write_toml(data, file_path):
    with open(file_path, "w") as file:
        toml.dump(data, file)


import os
import zhipuai


class ChatZhipu(object):
    def __init__(self, zhipu_apikey="", model="chatglm_turbo", temperature=0.95, top_p=0.7, incremental=True):
        zhipuai.api_key = os.environ.get("zhipu_apikey", zhipu_apikey)
        self.model = model
        self.temperature = temperature
        self.top_p = top_p
        self.incremental = incremental

    def invoke(self, query, **kwargs):
        response = zhipuai.model_api.invoke(
            prompt=[{"role": "user", "content": query}],
            model=self.model,
            temperature=self.temperature,
            top_p=self.top_p,
            incremental=self.incremental,
            **kwargs
        )
        return response['data']['choices'][0]['content']

    def stream(self, query, **kwargs):
        response = zhipuai.model_api.sse_invoke(
            prompt=[{"role": "user", "content": query}],
            model=self.model,
            temperature=self.temperature,
            top_p=self.top_p,
            incremental=self.incremental,
            **kwargs
        )
        '''
          说明：
          add: 事件流开启
          error: 平台服务或者模型异常，响应的异常事件
          interrupted: 中断事件，例如：触发敏感词
          finish: 数据接收完毕，关闭事件流
        '''
        for event in response.events():
            if event.event == "add":
                yield event.data
            elif event.event == "error" or event.event == "interrupted":
                yield event.data
            elif event.event == "finish":
                yield event.data
                yield event.meta
            else:
                yield event.data




from pylmkit.llms import EmbeddingsHuggingFaceBge
from pylmkit.llms import EmbeddingsHuggingFace
from pylmkit.perception.text import Dict2Document, Text2Document
from typing import Any, Iterable, List, Optional, Tuple, Type
from pylmkit.perception.text import DocumentLoader, WebLoader
from pylmkit.app import RolePlay


class VectorDB(object):
    def __init__(self, corpus=None, embed_model=None, vdb_model=None, init_vdb=None):
        self.vdb = init_vdb
        if corpus and embed_model and vdb_model:
            corpus = self.any2doc(corpus)
            self.vdb = vdb_model.from_documents(corpus, embed_model, ids=[i for i in range(1, len(corpus) + 1)])

    @classmethod
    def load(cls, vdb_model, embed_model, vdb_path, vdb_name="index", is_return=True, extend=True, **kwargs):
        _vdb = vdb_model.load_local(vdb_path, embed_model, index_name=vdb_name, **kwargs)
        cls()._base(_vdb, is_return=is_return, extend=extend)

    @classmethod
    def save(cls, vdb_path: str, vdb_name: str = "index", vdb_model=None):
        if vdb_model is None:
            vdb_model = cls().vdb
        vdb_model.save_local(folder_path=vdb_path, index_name=vdb_name)

    @classmethod
    def add(cls, corpus, vdb_model=None, is_return=True, extend=True):
        if vdb_model is None:
            vdb_model = cls().vdb
        corpus = cls().any2doc(corpus)
        vdb = vdb_model.add_documents(documents=corpus)
        cls()._base(vdb, is_return=is_return, extend=extend)

    @classmethod
    def update(cls, ids, corpus, vdb_model=None, is_return=True, extend=True):
        if vdb_model is None:
            vdb_model = cls().vdb
        corpus = cls().any2doc(corpus)
        vdb_model.update_documents(ids=ids, documents=corpus)
        cls()._base(vdb_model, is_return=is_return, extend=extend)

    def get(self, ids, vdb_model=None):
        if vdb_model is None:
            vdb_model = self.vdb
        return vdb_model._collection.get(ids=ids)

    def delete(self, ids, vdb_model=None, is_return=False, extend=False):
        if vdb_model is None:
            vdb_model = self.vdb
        vdb_model = vdb_model._collection.delete(ids=ids)
        self._base(vdb_model, is_return=is_return, extend=extend)

    def count(self, vdb_model=None):
        if vdb_model is None:
            vdb_model = self.vdb
        return vdb_model._collection.count()

    def any2doc(self, corpus):
        # any, str dict doc
        if corpus and isinstance(corpus[0], str):
            corpus = Text2Document.get(texts=corpus, is_return=True, return_mode='doc', extend=False)
        elif corpus and isinstance(corpus[0], dict):
            corpus = Dict2Document.get(doc_dict=corpus, is_return=True, extend=False, return_mode='doc')
        else:
            corpus = corpus
        return corpus

    def _base(self, vdb_model, is_return=True, extend=False):
        if extend:
            self.vdb = vdb_model
        if is_return:
            return vdb_model

    def ra(self,
           query: str,
           topk: int = 5,
           search_language=[],
           lambda_val: float = 0.025,
           filter: Optional[str] = None,
           n_sentence_context: int = 2,
           **kwargs: Any, ):
        return self.vdb.similarity_search(
            query=query,
            k=topk,
            lambda_val=lambda_val,
            filter=filter,
            n_sentence_context=n_sentence_context,
            **kwargs
        )

    def retriever(self,
                  topk,
                  filter_metadata={},
                  fetch_k=20,
                  lambda_mult=0.5,
                  score_threshold=0.8,
                  search_type="similarity"
                  ):
        """from langchain: Return VectorStoreRetriever initialized from this VectorStore.

                Args:
                    search_type (Optional[str]): Defines the type of search that
                        the Retriever should perform.
                        Can be "similarity" (default), "mmr", or "similarity_score_threshold".
                    search_kwargs (Optional[Dict]): Keyword arguments to pass to the
                        search function. Can include things like:

                            public setting:
                                k: Amount of documents to return (Default: 4)
                                filter: Filter by document metadata

                            similarity_score_threshold setting:
                                score_threshold: Minimum relevance threshold
                                    for similarity_score_threshold

                            mmr setting:
                                fetch_k: Amount of documents to pass to MMR algorithm (Default: 20)
                                lambda_mult: Diversity of results returned by MMR;
                                    1 for minimum diversity and 0 for maximum. (Default: 0.5)


                Returns:
                    VectorStoreRetriever: Retriever class for VectorStore.

                Examples:

                .. code-block:: python

                    # Retrieve more documents with higher diversity
                    # Useful if your dataset has many similar documents
                    docsearch.as_retriever(
                        search_type="mmr",
                        search_kwargs={'k': 6, 'lambda_mult': 0.25}
                    )

                    # Fetch more documents for the MMR algorithm to consider
                    # But only return the top 5
                    docsearch.as_retriever(
                        search_type="mmr",
                        search_kwargs={'k': 5, 'fetch_k': 50}
                    )

                    # Only retrieve documents that have a relevance score
                    # Above a certain threshold
                    docsearch.as_retriever(
                        search_type="similarity_score_threshold",
                        search_kwargs={'score_threshold': 0.8}
                    )

                    # Only get the single most similar document from the dataset
                    docsearch.as_retriever(search_kwargs={'k': 1})

                    # Use a filter to only retrieve documents from a specific paper
                    docsearch.as_retriever(
                        search_kwargs={'filter': {'paper_title':'GPT-4 Technical Report'}}
                    )
                """
        search_kwargs = {}
        for kwarg in [topk, filter_metadata, fetch_k, lambda_mult, score_threshold]:
            if kwarg:
                search_kwargs[str(kwarg)] = kwarg
        return self.vdb.as_retriever(search_kwargs=search_kwargs, search_type=search_type)


class BaseRAG(VectorDB, RolePlay):
    def __init__(self,
                 embed_model,
                 vdb_model,
                 llm_model,
                 corpus=None,
                 vdb_path=None,
                 vdb_name="index",
                 init_vdb=None,
                 role_template="",
                 memory=None,
                 online_search_kwargs={},
                 return_language="English",
                 ):
        VectorDB.__init__(self, init_vdb=None)
        RolePlay.__init__(self,
                          role_template=role_template,
                          llm_model=llm_model,
                          memory=memory,
                          online_search_kwargs=online_search_kwargs,
                          return_language=return_language
                          )
        if init_vdb:
            VectorDB.__init__(self, init_vdb=init_vdb)
        elif vdb_path and vdb_name:
            super().load(vdb_model=vdb_model,
                         embed_model=embed_model,
                         vdb_path=vdb_path,
                         vdb_name=vdb_name,
                         is_return=False,
                         extend=True
                         )
        elif corpus:
            super().__init__(corpus=corpus, embed_model=embed_model, vdb_model=vdb_model)
        else:
            raise Exception("`corpus`, `vdb_path`, `init_vdb` Cannot all be None!")

    def invoke(
            self,
            query: str,
            topk: int = 5,
            search_language=[],
            lambda_val: float = 0.025,
            filter: Optional[str] = None,
            n_sentence_context: int = 2,
            ra_kwargs={},
            **kwargs
    ):
        ra_documents = super().ra(
            query=query,
            topk=topk,
            search_language=search_language,
            lambda_val=lambda_val,
            filter=filter,
            n_sentence_context=n_sentence_context,
            **ra_kwargs
        )
        return super().invoke(query=query, ra_documents=ra_documents, **kwargs)

    def stream(
            self,
            query: str,
            topk: int = 5,
            search_language=[],
            lambda_val: float = 0.025,
            filter: Optional[str] = None,
            n_sentence_context: int = 2,
            ra_kwargs={},
            **kwargs
    ):
        ra_documents = super().ra(
            query=query,
            topk=topk,
            search_language=search_language,
            lambda_val=lambda_val,
            filter=filter,
            n_sentence_context=n_sentence_context,
            **ra_kwargs
        )
        return super().stream(query=query, ra_documents=ra_documents, **kwargs)


class DocRAG(BaseRAG):
    def __init__(self,
                 embed_model,
                 vdb_model,
                 llm_model,
                 corpus=None,
                 vdb_path=None,
                 vdb_name="index",
                 init_vdb=None,
                 role_template="",
                 memory=None,
                 online_search_kwargs={},
                 return_language="English",
                 ):
        super().__init__(
            embed_model=embed_model,
            vdb_model=vdb_model,
            llm_model=llm_model,
            corpus=corpus,
            vdb_path=vdb_path,
            vdb_name=vdb_name,
            init_vdb=init_vdb,
            role_template=role_template,
            memory=memory,
            online_search_kwargs=online_search_kwargs,
            return_language=return_language
        )


class WebRAG(BaseRAG):
    def __init__(self,
                 embed_model,
                 vdb_model,
                 llm_model,
                 corpus=None,
                 vdb_path=None,
                 vdb_name="index",
                 init_vdb=None,
                 role_template="",
                 memory=None,
                 online_search_kwargs={},
                 return_language="English",
                 ):
        super().__init__(
            embed_model=embed_model,
            vdb_model=vdb_model,
            llm_model=llm_model,
            corpus=corpus,
            vdb_path=vdb_path,
            vdb_name=vdb_name,
            init_vdb=init_vdb,
            role_template=role_template,
            memory=memory,
            online_search_kwargs=online_search_kwargs,
            return_language=return_language
        )



